program Ultimatch_Huber

  use sep
  use helix
  use nhelix
  use nhelicon
  use nchelicon
  use cartesian
  use findmatch_Huber_B
  use verbosity_mod

  use huber_reg_mod
  use lapfac

  !  use irls_priv

  !  use build_filter_mod

  implicit none

  integer, external                    :: getch
  real,          pointer, dimension(:) :: data1,data2,data3,out
  !  type (nfilter), pointer              :: sig_pef
  type (filter)                        :: match,laplac
  type (filter), pointer, dimension(:) :: nmatch
  type (nfilter)                       :: nsmatch,rough

  ! Huber stuff
  real                                 :: eps,gamd,gamx
  integer                              :: l1,nmodel

  integer :: mode               ! 1=single filter, 2=patches, 3=NS filtering
  integer :: ndim               ! dimension of dataset
  integer :: niter,i,ilag,idim,ipatch,ipanels,npanels

  logical :: twodim,interleave 
  character (len=20) :: targetname
  integer, dimension(:), pointer :: nfilt  ! no. of coefficients in each filter
  integer, dimension(:), pointer :: n      ! data-size, dimension(ndim)
  real,    dimension(:), pointer :: d, o
  integer, dimension(:), pointer :: npatch ! no. of patches, dimension(ndim)
  integer, dimension(:), pointer :: nwind  ! overlapping patch-size
  integer, dimension(:), pointer :: idx    ! index counter (ndim)

  integer, dimension(:), pointer :: nh    ! length of each filter in the table
  integer, dimension(:), pointer :: pch   ! ns filter indexing
  integer, dimension(:), pointer :: psize ! non-overlapping patch-size

  call sep_init(SOURCE)
  ndim=sep_dimension()

  allocate(n(ndim), o(ndim), d(ndim))
  call from_history('n',n)
  call from_history('o',o)
  call from_history('d',d)
  interleave=.false.
  if (getch('target','s',targetname) == 0) interleave=.true.

  call from_param('2D',twodim,.false.)
  write(0,*) '2D',twodim
  npanels=1
  if (twodim) then
     allocate(nfilt(2))
     nfilt=n(1:2)
     npanels=1
     if (ndim>2) npanels=n(3)
     if (interleave) then 
        npanels=1
        if (ndim>3) npanels=n(4)
        call to_history('n3',npanels)
        call to_history('n4',1)
     end if
     deallocate(n)
     ndim=2
     allocate(n(2))
     n=nfilt
     deallocate(nfilt)
  end if

  allocate(nfilt(ndim),idx(ndim))
  nfilt=1; nfilt(1)=21

  call from_param('mode',mode,1)
  call from_param('verb',verb,10)
  call from_param('nfilt',nfilt,nfilt)
  call from_param("gamd",gamd,1.)
  call from_param("gamx",gamx,0.)
  call from_param('niter',niter,product(nfilt)-1)
  call from_param('eps',eps,1.) ! for the adaptive subtraction
 
  ! Patching parameters
  allocate(npatch(ndim),nwind(ndim))
  npatch=1
  call from_param('npatch',npatch,npatch)

  do i=1,ndim
     if (npatch(i)==1) then 
        nwind(i)=n(i)
     else 
        nwind(i)=2*n(i)/npatch(i)
     end if
  end do

  call from_param('nwind',nwind,nwind)

  allocate(data1(product(n)),data2(product(n)),data3(product(n)),out(product(n)))

  ! Create a single filter
  call allocatehelix(match,product(nfilt))
  do ilag=1,product(nfilt)
     call line2cart(nfilt,ilag,idx)
     match%lag(ilag)=idx(1)-nfilt(1)/2-1
     do idim=2,ndim
        match%lag(ilag)=match%lag(ilag)+(idx(idim)-nfilt(idim)/2-1)*product(n(1:idim-1))
     end do
  end do

  if (verb>0) write(0,*) '############################'
  if (verb>0) write(0,*) '#      Huber parameters     '
  if (verb>0) write(0,*) '#---------------------------'
  if (verb>0) write(0,*) '#    eps   = ',eps
  if (verb>0) write(0,*) '#    niter = ',niter
  if (verb>0) write(0,*) '#    gamd   = ',gamd
  if (verb>0) write(0,*) '#    gamx   = ',gamx
  if (verb>0) write(0,*) '############################'
  
  select case(mode)

  case(1)
     if (verb>0) write(0,*) '############################'
     if (verb>0) write(0,*) '# Single matched-filter...  '
     if (verb>0) write(0,*) '#---------------------------'
     if (verb>1) write(0,*) '#    nfilt  = ',nfilt
     if (verb>0) write(0,*) '############################'


  case(2)
     if (verb>0) write(0,*) '#####################################'
     if (verb>0) write(0,*) '# Non-stationary matched-filtering...'
     if (verb>0) write(0,*) '#------------------------------------'
     if (verb>1) write(0,*) '#    ndim   = ',ndim
     if (verb>1) write(0,*) '#    n      = ',n
     if (verb>1) write(0,*) '#    npatch = ',npatch
     if (verb>1) write(0,*) '#    nfilt  = ',nfilt

     !! Create a non-stationary filter
     allocate(nh(product(npatch)))      !! nh is the length of each filter in table
     nh=product(nfilt)                  !! Set them all constant
     call psize_init(n,npatch,psize)    !! psize(ndim) is the patch-size

     if (verb>0) write(0,*) '#    patch-size = ',psize
     if (verb>0) write(0,*) '#####################################'

     call pch_init(n,npatch,psize,pch)  !! pch(ndata) is the index
     call nallocate(nsmatch,nh,pch)

     do ipatch=1,product(npatch)
        nsmatch%hlx(ipatch)%lag=match%lag
     end do

     call laplac_init(laplac,nsmatch,npatch) !! laplac is the regularizer

  case default
     call seperr('mode not implemented')
  end select

  select case(mode)     
  case(1)     
     nmodel=product(nfilt)    
  case(2)    
     nmodel=size(nsmatch%hlx(1)%lag)*size(nsmatch%hlx)  
  case default
     call seperr('Mode not implemented')  
  end select
  if (verb>1) write(0,*) 'Size model space =', nmodel

  call huber_init_out(gamd,gamx,2,eps,nmodel,verb)
  
  panels:do ipanels=1,npanels

     if (verb>1) write(0,*) '###################################'
     if (verb>1) write(0,*) '   Panel = ',ipanels,' out of ',npanels
     if (verb>1) write(0,*) '-----------------------------------'

     out=0.

     ! Read in data

     call sep_read(data1)
     if (interleave) then
        call sep_read(data2)
     else
        call sep_read(data2,'target')
     end if

     call huber_init_in(data2)

     if ((maxval(data1).eq.minval(data1)).or.(maxval(data2).eq.minval(data2))) then
        call sep_write(out)
        cycle
     end if
     select case(mode)

     case(1) ! One filter

        if (verb>2) write(0,*) ' ----------------------------'
        if (verb>2) write(0,*) ' Calling findmatch with Huber'
        if (verb>2) write(0,*) ' ----------------------------'
        
        call findmatch(data1,data2,match,niter)

        call dumpfilter(match)
        call applymatch(data1,out,match)

     case(2) ! NS filters
        
        if (verb>2) write(0,*) ' -------------------------------'
        if (verb>2) write(0,*) ' Calling findnsmatch with Huber '
        if (verb>2) write(0,*) ' -------------------------------'
        
        call findnsmatch(data1,data2,nsmatch,laplac,niter) 
        
        if (verb>2) write(0,*) ' Calling applynsmatch...'
        call applynsmatch(data1,out,nsmatch) 

     case default
        call seperr('Mode not implemented')  
     end select

     call huber_close()

     call sep_write(out)

  end do panels

  write(0,*) '---- The End ----'

  call sep_close()

contains

  !! Calculate non-overlapping patch-size given a certain number of patches
  subroutine psize_init(n,npatch,psize)
    integer, dimension(:), intent(in)  :: n,npatch
    integer, dimension(:), pointer     :: psize
    allocate(psize(ndim))
    do i=1,ndim
       psize(i)=1+(n(i)-1)/npatch(i)
    end do
  end subroutine psize_init

  !! Divides data-space between rectangular patches
  subroutine pch_init(n,npatch,psize,pch)
    integer, dimension(:), intent(in)  :: n,npatch,psize
    integer, dimension(:), pointer     :: pch,ii
    integer :: i,j

    allocate(pch(product(n)),ii(size(n)))
    do i=1,product(n)
       pch(i)=0
       call line2cart(n,i,ii)
       do j=1,size(n)
          ii(j)=(1+(ii(j)-1)/psize(j))
       end do
       call cart2line(npatch,ii,pch(i))
    end do
    deallocate(ii)
  end subroutine pch_init

  !! Better roughener 
  subroutine create_lap_2d_rough(rough,ncoef,npatch)
    
    type(nfilter)        :: rough
    integer              :: ncoef,npatch(2)
    integer              :: npts(2)
    integer, allocatable :: nh(:),map(:)
    
    npts=(/npatch(1),npatch(2)/)
    allocate(nh(product(npts)))
    allocate(map(ncoef*product(npts)))
    
    nh=ncoef
    map=1
    
    nh=min(6,npts(1))*2
    call nallocate(rough,nh,map)
    rough%hlx(1)=lapfac2(.01,npts(1),nh(1)/2)
    
    deallocate(nh,map)
    
  end subroutine create_lap_2d_rough

  !! Create a roughener to dampen NS filter estimation
  subroutine laplac_init(laplac,nsfilt,npatch)
    type  (filter)        :: laplac
    type (nfilter)        :: nsfilt
    integer, dimension(:) :: npatch

    real    :: e1,e2
    integer :: nhmax
    nhmax=size(nsfilt%hlx(1)%lag)

    select case(size(npatch))
    case (1)
       call allocatehelix(laplac,1)  
       laplac%lag(1)=nhmax;   laplac%flt(1)=-1.
    case (2)
       !! 1-D smoothing only simple gradient

       ! crap 2-D smoothing

       call allocatehelix(laplac,3)  
       laplac%lag=(/ nhmax, npatch(1)*nhmax, nhmax*(npatch(1)+1)  /)
       e1=0.; e2=0.
       laplac%flt=(/ e1-1.,      e2-1.,     (1.-e1)*(1.-e2)  /)

       if (verb>5) then
          write(0,*) '  Roughening filter...'
          call dumpfilter(laplac)
       end if

    case default
       write(0,*) 'Testmatch'
       call seperr('dimensionality not implemented')
    end select

  end subroutine laplac_init

end program Ultimatch_Huber













